#   Copyright Red Hat, Inc. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import time

import os

import cicoclient.client as client
import cicoclient.exceptions as exceptions


class CicoWrapper(client.CicoClient):
    """
    Wrapper library around the available API calls for the admin.ci.centos.org
    node infrastructure.
    """
    def __init__(self, **params):
        super(CicoWrapper, self).__init__(**params)
        self.user_agent = 'python-cicoclient-wrapper'
        self.api_key = self._lookup_api_key(params.get('api_key'))

        self._full_inventory = self._self_inventory = None

    @staticmethod
    def _lookup_api_key(api_key):
        if api_key is not None:
            return api_key

        try:
            return os.environ['CICO_API_KEY']
        except KeyError:
            pass

        for key_file in ('~/.duffy.key', '~/duffy.key'):
            try:
                return open(os.path.expanduser(key_file)).read().strip()
            except IOError:
                pass

        return None

    @property
    def full_inventory(self):
        """
        Returns a full inventory
        Some additional work required to provide consistent and consumable
        output.
        Inventory output only contains values, no keys - Add the keys to
        the output so that it can be consumed more easily.
        """
        if self._full_inventory:
            return self._full_inventory

        resp, inventory = self.get('Inventory')

        keys = ['host_id', 'hostname', 'ip_address', 'chassis',
                'used_count', 'current_state', 'comment', 'distro',
                'rel', 'centos_version', 'architecture', 'node_pool',
                'console_port', 'flavor']

        real_inventory = dict()
        for host in inventory:
            real_inventory[host[1]] = dict()
            for key in keys:
                real_inventory[host[1]][key] = host[keys.index(key)]

        self._full_inventory = real_inventory

        return self._full_inventory

    @property
    def self_inventory(self):
        """
        Inventory output will only contain the server name and the session ID
        when a key is provided. Provide the same format as with the full
        inventory instead for consistency.
        """
        if self.api_key is None:
            return {}

        if self._self_inventory:
            return self._self_inventory

        resp, self_inventory = self.get('Inventory?key=%s' % self.api_key)
        real_self_inventory = dict()

        for host in self_inventory:
            real_self_inventory[host[0]] = self.full_inventory[host[0]]

        self._self_inventory = real_self_inventory

        return self._self_inventory

    def _ssid_inventory(self, inventory, ssid):
        """
        Filters an inventory to only return servers matching ssid
        """
        matching_hosts = {}
        for host in inventory:
            if inventory[host]['comment'] == ssid:
                matching_hosts[host] = inventory[host]

        return matching_hosts

    def inventory(self, all=False, ssid=None):
        """
        Returns a node inventory. If an API key is specified, only the nodes
         provisioned by this key will be returned.

        :return: { inventory }
        """
        if all or self.api_key is None:
            if ssid is not None:
                return self._ssid_inventory(self.full_inventory, ssid)
            else:
                return self.full_inventory
        else:
            if ssid is not None:
                return self._ssid_inventory(self.self_inventory, ssid)
            else:
                return self.self_inventory

    def node_get(self, arch=None, ver=None, flavor=None, count=1,
                 retry_count=1, retry_interval=10):
        """
        Requests specified number of nodes with the provided parameters.

        :param arch: Server architecture (ex: x86_64)
        :param ver: CentOS version (ex: 7)
        :param count: Number of servers (ex: 2)
        :parma flavor: The flavor of machine to use (multi-arch only)
        :param retry_count: Number of times to retry in case of failure (ex: 5)
        :param retry_interval: Wait in seconds between each retry (ex: 30)
        :return: [ [ requested_hosts ], ssid ]
        """
        if self.api_key is None:
            raise exceptions.ApiKeyRequired

        args = "key=%s" % self.api_key
        if arch is not None:
            args += "&arch=%s" % arch
        if ver is not None:
            args += "&ver=%s" % ver
        if flavor is not None:
            args += "&flavor=%s" % flavor
        args += "&count=%s" % count

        resp, body = self.get('Node/get?%s' % args)
        if not body:
            for _ in range(retry_count):
                time.sleep(retry_interval)
                resp, body = self.get('Node/get?%s' % args)
                if body:
                    break

        if not body:
            raise exceptions.NoInventory

        # Get the hosts that were requested.
        # Note: We have to iterate over full inventory instead of just the
        # hosts we got back from the response because the reply contains the
        # fqdn of the host while the full inventory only contains a short name.
        requested_hosts = dict()
        for host in self.full_inventory:
            for full_host in body['hosts']:
                if host in full_host:
                    requested_hosts[host] = self.full_inventory[host]

        return requested_hosts, body['ssid']

    def node_done(self, ssid=None):
        """
        Release the servers for the specified ssid.
        The API doesn't provide any kind of output, try to be helpful by
        providing the list of servers to be released.

        :param ssid: ssid of the server pool
        :return: [ requested_hosts ]
        """
        if self.api_key is None:
            raise exceptions.ApiKeyRequired

        if ssid is None:
            raise exceptions.SsidRequired

        # There is no body replied in this call so at least get the hosts for
        # the specified ssid to return them.
        requested_hosts = dict()
        for host in self.self_inventory:
            if ssid == self.self_inventory[host]['comment']:
                requested_hosts[host] = self.full_inventory[host]

        args = "key={key}&ssid={ssid}".format(key=self.api_key, ssid=ssid)

        resp, body = self.get('Node/done?%s' % args)

        return requested_hosts



#   Copyright Red Hat, Inc. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import logging

from cliff.lister import Lister

from cicoclient import utils
from cicoclient.wrapper import CicoWrapper


class Inventory(Lister):
    """Returns a node inventory from the ci.centos.org infrastructure."""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Inventory, self).get_parser(prog_name)
        parser.add_argument(
            '--all',
            action='store_true',
            default=False,
            help='Display all nodes, regardless if an API key is used.'
        )
        parser.add_argument(
            '--ssid',
            metavar="<ssid>",
            default=None,
            help='Only return nodes matching the provided ssid.'
        )
        return parser

    @utils.log_method(log)
    def take_action(self, parsed_args):
        api = CicoWrapper(
            endpoint=self.app.options.endpoint,
            api_key=self.app.options.api_key
        )

        inventory = api.inventory(all=parsed_args.all,
                                  ssid=parsed_args.ssid)

        columns = ('host_id', 'hostname', 'ip_address', 'chassis',
                   'used_count', 'current_state', 'comment', 'distro',
                   'rel', 'centos_version', 'architecture', 'node_pool',
                   'console_port', 'flavor')

        return (columns,
                (utils.get_dict_properties(inventory[host], columns)
                 for host in inventory))


class NodeGet(Lister):
    """Requests nodes from the ci.centos.org infrastructure"""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(NodeGet, self).get_parser(prog_name)
        parser.add_argument(
            '--arch',
            metavar='<arch>',
            choices=['i386', 'x86_64', 'aarch64', 'ppc64le'],
            default='x86_64',
            help='Requested server architecture. Defaults to x86_64.'
        )
        parser.add_argument(
            '--release',
            metavar='<release>',
            choices=['5', '6', '7', '8', '8-stream','9-stream'],
            default='7',
            help='Requested CentOS release. Defaults to 7.'
        )
        parser.add_argument(
            '--count',
            metavar='<count>',
            type=int,
            default=1,
            help='Requested number of servers. Defaults to 1.'
        )
        parser.add_argument(
            '--retry-count',
            metavar='<count>',
            type=int,
            default=1,
            help='Number of retries to do in case of failure. Defaults to 1.'
        )
        parser.add_argument(
            '--retry-interval',
            metavar='<seconds>',
            type=int,
            default=10,
            help='Wait between subsequent retries. Defaults to 10 (seconds).'
        )
        parser.add_argument(
            '--flavor',
            metavar='<flavor>',
            default=None,
            choices=['tiny', 'small', 'medium', 'lram.tiny', 'lram.small',
                     'lram.medium', 'xram.tiny', 'xram.small',
                     'xram.medium', 'xram.large'],
            help='The flavor of the node.'
        )
        return parser

    @utils.log_method(log)
    def take_action(self, parsed_args):
        api = CicoWrapper(
            endpoint=self.app.options.endpoint,
            api_key=self.app.options.api_key
        )

        hosts, ssid = api.node_get(arch=parsed_args.arch,
                                   ver=parsed_args.release,
                                   count=parsed_args.count,
                                   retry_count=parsed_args.retry_count,
                                   retry_interval=parsed_args.retry_interval,
                                   flavor=parsed_args.flavor)

        columns = ('host_id', 'hostname', 'ip_address', 'chassis',
                   'used_count', 'current_state', 'comment', 'distro',
                   'rel', 'centos_version', 'architecture', 'node_pool',
                   'console_port', 'flavor')

        return (columns,
                (utils.get_dict_properties(hosts[host], columns)
                 for host in hosts))


class NodeDone(Lister):
    """Releases nodes from the ci.centos.org infrastructure for a ssid"""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(NodeDone, self).get_parser(prog_name)
        parser.add_argument(
            'ssid',
            metavar='<ssid>',
            help='SSID of the server pool to release'
        )
        return parser

    @utils.log_method(log)
    def take_action(self, parsed_args):
        api = CicoWrapper(
            endpoint=self.app.options.endpoint,
            api_key=self.app.options.api_key
        )

        hosts = api.node_done(ssid=parsed_args.ssid)

        columns = ('host_id', 'hostname', 'ip_address', 'chassis',
                   'used_count', 'current_state', 'comment', 'distro',
                   'rel', 'centos_version', 'architecture', 'node_pool',
                   'console_port', 'flavor')

        return (columns,
                (utils.get_dict_properties(hosts[host], columns)
                 for host in hosts))



#!/usr/bin/python
#   Copyright Red Hat, Inc. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from ansible.module_utils.basic import AnsibleModule

DOCUMENTATION = '''
---
module: cico
short_description: Ansible module to manage ci.centos.org node lifecycle
version_added: "2.0"
author: "David Moreau Simard <dms@redhat.com>"
description:
    - Ansible module to manage ci.centos.org node lifecycle
options:
    action:
        description:
            - Action to take
        choices: [get, done, list]
        required: true
    arch:
        description:
            - Server architecture
        choices: [i386, x86_64, aarch64, ppc64le]
        default: x86_64
    release:
        description:
            - CentOS release
        choices: [5, 6, 7, 8, 8-stream, 9-stream]
        default: 7
    flavor:
        description:
            - The flavor (size) of an altarch Node
        choices:
            - tiny
            - small
            - medium
            - lram.tiny
            - lram.small
            - lram.medium
            - xram.tiny
            - xram.small
            - xram.medium
            - xram.large
        default: small

    count:
        description:
            - Number of nodes
        default: 1
    retry_count:
        description:
            - Number of retries to do in case of failure.
        default: 1
    retry_interval:
        description:
            - Wait (in seconds) between subsequent retries.
        default: 10
    endpoint:
        description:
            - API endpoint
        default: http://admin.ci.centos.org:8080/
    api_key:
        description:
            - API key
        default: >
            CICO_API_KEY environment variable, the contents of ~/.duffy.key,
            the contents of ~/duffy.key, or None
    ssid:
        description:
            - SessionID, required with action 'done', optional with 'list'.

requirements:
    - "python >= 2.6"
    - "python-cicoclient"
'''

EXAMPLES = '''
# Retrieve full inventory
- cico:
    action: list
  register: data

# Retrieve inventory tied to API key
- cico:
    action: list
    api_key:  723ef3ce-4ea4-4e8d-9c8a-20a8249b2955
  register: data

# Retrieve inventory tied to a SSID
- cico:
    action: list
    ssid:  3e03553f-ae28-4a68-b879-f0fdbf949d5d
  register: data

# Request one CentOS 7 x86_64 node
- cico:
    action: get
    api_key: 723ef3ce-4ea4-4e8d-9c8a-20a8249b2955
  register: data

# Request one CentOS 7 x86_64 node with increased tolerance failure
- cico:
    action: get
    api_key: 723ef3ce-4ea4-4e8d-9c8a-20a8249b2955
    retry_count: 3
    retry_interval: 60
  register: data

# Request two CentOS 6 i386 nodes
- cico:
    action: get
    api_key: 723ef3ce-4ea4-4e8d-9c8a-20a8249b2955
    arch: i386
    release: 6
    count: 2
  register: data

# Release nodes requested in a registered 'get' action
- cico:
    action: done
    api_key: 723ef3ce-4ea4-4e8d-9c8a-20a8249b2955
    ssid: data.ssid

# Release nodes for a specific ssid
- cico:
    action: done
    api_key: 723ef3ce-4ea4-4e8d-9c8a-20a8249b2955
    ssid: 3e03553f-ae28-4a68-b879-f0fdbf949d5d
'''

try:
    from cicoclient.wrapper import CicoWrapper
    HAS_CICO = True
except ImportError:
    HAS_CICO = False


def main():
    argument_spec = dict(
        action=dict(required=True, choices=['get', 'done', 'list']),
        arch=dict(default='x86_64', choices=['i386', 'x86_64', 'aarch64',
                                             'ppc64le']),
        flavor=dict(default=None, choices=['tiny', 'small', 'medium',
                                              'lram.tiny', 'lram.small',
                                              'xram.tiny', 'xram.small',
                                              'xram.medium', 'xram.large']),
        release=dict(default='7', choices=['5', '6', '7', '8', '8-stream', '9-stream']),
        count=dict(default=1, type='int'),
        retry_count=dict(default=1, type='int'),
        retry_interval=dict(default=10, type='int'),
        endpoint=dict(default='http://admin.ci.centos.org:8080/'),
        api_key=dict(default=None, no_log=True),
        ssid=dict(default=None),
    )
    module = AnsibleModule(argument_spec)

    if not HAS_CICO:
        module.fail_json(msg='cicoclient is required for this module.')

    action = module.params['action']
    arch = module.params['arch']
    release = module.params['release']
    count = module.params['count']
    retry_count = module.params['retry_count']
    retry_interval = module.params['retry_interval']
    endpoint = module.params['endpoint']
    api_key = module.params['api_key']
    ssid = module.params['ssid']
    flavor = module.params['flavor']

    if action == 'done' and ssid is None:
        module.fail_json(msg='A SSID is required when releasing nodes.')

    try:
        api = CicoWrapper(
            endpoint=endpoint,
            api_key=api_key
        )

        if api.api_key is None:
            module.fail_json(msg='An API key is required for this module.')

        if action == 'get':
            hosts, new_ssid = api.node_get(arch=arch, ver=release, count=count,
                                           retry_count=retry_count,
                                           retry_interval=retry_interval,
                                           flavor=flavor)
            data = {
                'message': 'Requested servers successfully',
                'hosts': hosts,
                'ssid': new_ssid
            }
            module.exit_json(changed=True, results=data)

        if action == 'done':
            hosts = api.node_done(ssid=ssid)
            data = {
                'message': 'Released servers successfully',
                'hosts': hosts
            }
            module.exit_json(changed=True, results=data)

        if action == 'list':
            hosts = api.inventory(ssid=ssid)

            data = {
                'message': 'Listed servers successfully',
                'hosts': hosts
            }
            module.exit_json(changed=True, results=data)

    except Exception as e:
        module.fail_json(msg=e.message)


if __name__ == '__main__':
    main()



#   Copyright Red Hat, Inc. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import sys

import cicoclient
from cliff.app import App
from cliff.commandmanager import CommandManager


class CicoCli(App):
    """
    CLI interface boilerplate with cliff
    """
    def __init__(self):
        super(CicoCli, self).__init__(
            description='CLI interface to admin.ci.centos.org',
            version=cicoclient.__version__,
            command_manager=CommandManager('cico.cli'),
            deferred_help=True,
            )

    def build_option_parser(self, description, version):
        parser = super(CicoCli, self).build_option_parser(description, version)

        # Global arguments
        parser.add_argument(
            '--endpoint',
            metavar='<endpoint>',
            help='Endpoint to the admin.ci.centos.org service.\n'
                 ' Defaults to: http://admin.ci.centos.org:8080/',
            default='http://admin.ci.centos.org:8080/'
        )
        parser.add_argument(
            '--api-key',
            metavar='<api-key>',
            help='API key to admin.ci.centos.org service. If not provided the'
                 ' value of the CICO_API_KEY environment variable will be used'
                 ' if defined, followed by the contents of ~/.duffy.key if'
                 ' present, finally the contents of ~/duffy.key if present.',
            default=None
        )

        return parser

    def initialize_app(self, argv):
        self.LOG.debug('initialize_app')

    def prepare_to_run_command(self, cmd):
        self.LOG.debug('prepare_to_run_command %s', cmd.__class__.__name__)

    def clean_up(self, cmd, result, err):
        self.LOG.debug('clean_up %s', cmd.__class__.__name__)
        if err:
            self.LOG.debug('got an error: %s', err)


def main(argv=sys.argv[1:]):
    cicocli = CicoCli()
    return cicocli.run(argv)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#   Copyright Red Hat, Inc. All Rights Reserved.
#   Copyright 2015 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import six
import logging


def get_dict_properties(item, fields, mixed_case_fields=[], formatters={}):
    """Return a tuple containing the item properties.
    :param item: a single dict resource
    :param fields: tuple of strings with the desired field names
    :param mixed_case_fields: tuple of field names to preserve case
    :param formatters: dictionary mapping field names to callables
       to format the values
    """
    row = []

    for field in fields:
        if field in mixed_case_fields:
            field_name = field.replace(' ', '_')
        else:
            field_name = field.lower().replace(' ', '_')
        data = item[field_name] if field_name in item else ''
        if field in formatters:
            row.append(formatters[field](data))
        else:
            row.append(data)
    return tuple(row)


def log_method(log, level=logging.DEBUG):
    """Logs a method and its arguments when entered."""

    def decorator(func):
        func_name = func.__name__

        @six.wraps(func)
        def wrapper(self, *args, **kwargs):
            if log.isEnabledFor(level):
                pretty_args = []
                if args:
                    pretty_args.extend(str(a) for a in args)
                if kwargs:
                    pretty_args.extend(
                        "%s=%s" % (k, v) for k, v in six.iteritems(kwargs))
                log.log(level, "%s(%s)", func_name, ", ".join(pretty_args))
            return func(self, *args, **kwargs)

        return wrapper

    return decorator



#   Copyright Red Hat, Inc. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import logging
import requests
import json


class CicoClient(object):
    def __init__(self, **params):
        """
        Initialize the class, get the necessary parameters
        """
        self.user_agent = 'python-cicoclient'

        self.params = params
        self.log = self.log_wrapper()

        self.log.debug("Params: {0}".format(str(self.params)))

        self.endpoint = self.params['endpoint']
        if 'timeout' not in self.params:
            self.timeout = None

        self.http = requests.Session()

    def _request(self, url, method, **kwargs):
        if self.timeout is not None:
            kwargs.setdefault('timeout', self.timeout)

        kwargs.setdefault('headers', kwargs.get('headers', {}))
        kwargs['headers']['User-Agent'] = self.user_agent
        kwargs['headers']['Accept'] = 'application/json'
        kwargs['headers']['Content-Type'] = 'application/json'

        self.log.debug("{0} URL: {1}{2} - {3}"
                       .format(method, self.endpoint, url, str(kwargs)))

        resp = self.http.request(method, self.endpoint + url, **kwargs)

        body = None
        if resp.text:
            try:
                body = json.loads(resp.text)
            except ValueError:
                pass

        return resp, body

    def get(self, url, **kwargs):
        return self._request(url, 'GET', **kwargs)

    def log_wrapper(self):
        """
        Wrapper to set logging parameters for output
        """
        log = logging.getLogger('client.py')

        # Set the log format and log level
        try:
            if self.params['debug']:
                log.setLevel(logging.DEBUG)
        except KeyError:
            log.setLevel(logging.INFO)

        # Set the log format.
        stream = logging.StreamHandler()
        logformat = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
            datefmt='%b %d %H:%M:%S')
        stream.setFormatter(logformat)

        log.addHandler(stream)
        return log
